from typing import Mapping, Set, MutableMapping, Optional

import attr

__all__ = [
    "DependencyNode",
    "DependencyGraph",
]


@attr.define
class DependencyNode:
    name: str = attr.field()
    parents: Set["DependencyNode"] = attr.field(factory=set)
    children: Set["DependencyNode"] = attr.field(factory=set)

    def add_child(self, child: "DependencyNode"):
        self.children.add(child)
        child.parents.add(self)

    def get_children_closure(self, closure: Optional[Set[str]] = None) -> Set[str]:
        is_root = closure is None
        if closure is None:
            closure = set()
        if self.name in closure:
            return closure
        closure.add(self.name)
        for child in self.children:
            closure |= child.get_children_closure(closure)

        if is_root:
            closure.remove(self.name)
        return closure

    def get_parents_closure(self, closure: Optional[Set[str]] = None) -> Set[str]:
        is_root = closure is None
        if closure is None:
            closure = set()
        if self.name in closure:
            return closure
        closure.add(self.name)
        for parent in self.parents:
            closure |= parent.get_parents_closure(closure)

        if is_root:
            closure.remove(self.name)
        return closure

    def __hash__(self):
        return hash(self.name)


@attr.define
class DependencyGraph:
    nodes: MutableMapping[str, DependencyNode] = attr.field(factory=dict)

    def add_node(self, name: str) -> DependencyNode:
        if name not in self.nodes:
            self.nodes[name] = DependencyNode(name)
        return self.nodes[name]

    def add_edge(self, parent: str, child: str):
        parent_node = self.add_node(parent)
        child_node = self.add_node(child)
        parent_node.add_child(child_node)

    @classmethod
    def from_child_mapping(cls, mapping: Mapping[str, Set[str]]) -> "DependencyGraph":
        graph = cls()
        for var, children in mapping.items():
            graph.add_node(var)
            for child in children:
                graph.add_edge(var, child)
        return graph

    @classmethod
    def from_parent_mapping(cls, mapping: Mapping[str, Set[str]]) -> "DependencyGraph":
        graph = cls()
        for var, parents in mapping.items():
            graph.add_node(var)
            for parent in parents:
                graph.add_edge(parent, var)
        return graph

    def child_mapping(self) -> Mapping[str, Set[str]]:
        return {
            node.name: {child.name for child in node.children}
            for node in self.nodes.values()
        }

    def child_closure_mapping(self) -> Mapping[str, Set[str]]:
        return {node.name: node.get_children_closure() for node in self.nodes.values()}

    def parent_mapping(self) -> Mapping[str, Set[str]]:
        return {
            node.name: {parent.name for parent in node.parents}
            for node in self.nodes.values()
        }

    def parent_closure_mapping(self) -> Mapping[str, Set[str]]:
        return {node.name: node.get_parents_closure() for node in self.nodes.values()}



__all__ = ["adjust", "convert", "calculate_specific_humidity"]

import xarray as xr
import numpy as np

# Define constants
T_0 = 273.15                # Ice point temperature (Kelvins)
T_100 = T_0+100             # Steam point temperature (Kelvins)
ews = 1013.246              # Saturation vapour pressure at steam point temperature (normal atmosphere) (hPa)
ei0 = 6.1071                # Saturation vapour pressure at ice melting point temperature (normal atmosphere) (hPa)
eps=0.622                   # Ratio of molar masses of vapor and dry air

def adjust(rh: xr.DataArray,
           t: xr.DataArray
) -> xr.DataArray:
    """Correct relative humidity so that values are given with respect to
    saturation over ice in subfreezing conditions, and with respect to
    saturation over water (as given by the instrument) above the melting
    point temperature. Saturation water vapors are calculated after
    Groff & Gratch method.

    Parameters
    ----------
    rh : xr.DataArray
        Relative humidity
    t : xr.DataArray
        Air temperature

    Returns
    -------
    rh_wrt_ice_or_water : xr.DataArray
        CAdjusted relative humidity
    """
    # Convert to hPa (Groff & Gratch)
    e_s_wtr = 10**(-7.90298 * (T_100 / (t + T_0) - 1)
                   + 5.02808 * np.log10(T_100 / (t + T_0))
                   - 1.3816E-7 * (10**(11.344 * (1 - (t + T_0) / T_100)) - 1)
                   + 8.1328E-3 * (10**(-3.49149 * (T_100/(t + T_0) - 1)) -1)
                   + np.log10(ews))
    e_s_ice = 10**(-9.09718 * (T_0 / (t + T_0) - 1)
                   - 3.56654 * np.log10(T_0 / (t + T_0))
                   + 0.876793 * (1 - (t + T_0) / T_0)
                   + np.log10(ei0))

    # Define freezing point. Why > -100?
    nan_mask = t.notnull()
    freezing = (t < 0) & (t > -100) & nan_mask

    # Set to Groff & Gratch values when freezing, otherwise just rh
    rh_wrt_ice_or_water = rh.where(~freezing & nan_mask,
                                   other=rh*(e_s_wtr/e_s_ice))
    return rh_wrt_ice_or_water


def calculate_specific_humidity(t, p, rh_wrt_ice_or_water):
    """Calculate specific humidity.

    Parameters
    ----------
    t : xr.DataArray
        Air temperature
    p : xr.DataArray
        Air pressure
    rh_wrt_ice_or_water : xr.DataArray
        Adjusted relative humidity

    Returns
    -------
    xr.DataArray
        Specific humidity (kg/kg)
    """
    # Saturation vapour pressure above 0 C (hPa)
    es_wtr = 10**(-7.90298 * (T_100 / (t + T_0) - 1) + 5.02808 * np.log10(T_100 / (t + T_0))
                  - 1.3816E-7 * (10**(11.344 * (1 - (t + T_0) / T_100)) - 1)
                  + 8.1328E-3 * (10**(-3.49149 * (T_100 / (t + T_0) -1)) - 1) + np.log10(ews))

    # Saturation vapour pressure below 0 C (hPa)
    es_ice = 10**(-9.09718 * (T_0 / (t + T_0) - 1) - 3.56654
                  * np.log10(T_0 / (t + T_0)) + 0.876793
                  * (1 - (t + T_0) / T_0)
                  + np.log10(ei0))

    # Specific humidity at saturation (incorrect below melting point)
    q_sat = eps * es_wtr / (p - (1 - eps) * es_wtr)

    # Replace saturation specific humidity values below melting point
    freezing = t < 0
    q_sat[freezing] = eps * es_ice[freezing] / (p[freezing] - (1 - eps) * es_ice[freezing])

    # Mask where temperature or pressure are null values
    q_nan = np.isnan(t) | np.isnan(p)
    q_sat[q_nan] = np.nan

    # Convert to kg/kg
    return rh_wrt_ice_or_water * q_sat / 100

def convert(qh: xr.DataArray
) -> xr.DataArray:
    """Convert specific humidity from kg/kg to g/kg units

    Parameters
    ----------
    qh : xr.DataArray
        Specific humidity (kg/kg)

    Returns
    -------
    xr.DataArray
        Specific humidity (g/kg)
    """
    return 1000 * qh



import csv
from pathlib import Path
from typing import Dict, Union

import pandas as pd

DEFAULT_METADATA_PATH = (Path(__file__).parent / "file_attributes.csv").absolute()
DEFAULT_VARIABLES_PATH = (Path(__file__).parent / "variables.csv").absolute()
DEFAULT_VARIABLES_ALIASES_GCNET_PATH = (Path(__file__).parent / "variable_aliases_GC-Net.csv").absolute()
DEFAULT_VARIABLES_ALIASES_GLACIOBASIS_PATH = (Path(__file__).parent / "variable_aliases_GlacioBasis.csv").absolute()
DEFAULT_PAYLOAD_FORMATS_PATH = (Path(__file__).parent / "payload_formats.csv").absolute()
DEFAULT_PAYLOAD_TYPES_PATH = (Path(__file__).parent / "payload_types.csv").absolute()

def load_metadata(path: Union[None, str, Path] = None) -> Dict[str, str]:
    """
    Load metadata table from csv file
    """
    if path is None:
        path = DEFAULT_METADATA_PATH
    with open(path, "r") as f:
        csv_reader = csv.reader(f)
        return {row[0]: row[1] for row in csv_reader}


def load_variables(path: Union[None, str, Path] = None) -> pd.DataFrame:
    """
    Load variables table from csv file
    """
    if path is None:
        path = DEFAULT_VARIABLES_PATH
    return pd.read_csv(path, index_col=0, comment="#")



"""
Module for handling configuration loading and parsing of L0 data files.

This module provides the functionalities to interpret configuration files,
detect file types for data parsing, and process L0 data into xarray.Dataset
objects with associated metadata.

The module implements explicit input file type detection and parsing logic
for different data file types including `csv_v1`, `toa5`, and `csv_default`.
Additionally, it supports post-processing for time offsets and metadata
enrichment.

Functions
---------
- load_data_files: Reads and processes multiple data files given a configuration dictionary.
- load_config: Parses a TOML configuration file and produces a dictionary of configurations.
"""
import logging
import os
import re
from datetime import timedelta
from pathlib import Path
from typing import Dict, List, Optional, Sequence

import pandas as pd
import toml
import xarray as xr

from . import toa5

__all__ = [
    "load_data_files",
    "load_config",
]

logger = logging.getLogger(__name__)

DELIMITER = ","
COMMENT = "#"


# ---------------------------------------------------------------------
# Explicit input file type detection
# ---------------------------------------------------------------------


def _detect_file_type(conf: Dict) -> str:
    """Classify input file type explicitly.

    Returns one of:
      - 'csv_v1'      : legacy layout with year, doy, hhmm columns
      - 'toa5'        : Campbell Scientific TOA5
      - 'csv_default' : default CSV-like with timestamp in first column
    """
    infile = conf["file"]

    # 1) Respect explicit version hint from config
    file_version = conf.get("file_version", -1)
    if file_version == 1:
        return "csv_v1"

    # 2) Peek file header to detect TOA5
    try:
        with open(infile, "r", encoding="utf-8", errors="ignore") as f:
            # Read a handful of lines to detect markers
            header_lines = []
            for _ in range(10):
                line = f.readline()
                if not line:
                    break
                header_lines.append(line.strip())
    except Exception as e:
        logger.debug(f"Failed reading header for detection from {infile}: {e}")
        # Fall back to default if we cannot read
        return "csv_default"

    # Normalize: skip blank lines
    header_nonblank = [ln for ln in header_lines if ln]

    if header_nonblank:
        first = header_nonblank[0]

        # TOA5 files have a first line starting with 'TOA5'
        if re.match(r'^["]?TOA5', first):
            return "toa5"

    # Default CSV-like parser as a safe fallback
    return "csv_default"


def _parse_csv_v1(conf) -> pd.DataFrame:
    df = pd.read_csv(
        conf["file"],
        comment=COMMENT,
        parse_dates=True,
        na_values=conf["nodata"],
        names=conf["columns"],
        sep=DELIMITER,
        skiprows=conf["skiprows"],
        skip_blank_lines=True,
        usecols=range(len(conf["columns"])),
        low_memory=False,
    )
    df["time"] = pd.to_datetime(
        df.year.astype(str)
        + df.doy.astype(str).str.zfill(3)
        + df.hhmm.astype(str).str.zfill(4),
        format="%Y%j%H%M",
    )
    return df.set_index("time")


def _parse_csv_default(conf) -> pd.DataFrame:
    df = pd.read_csv(
        conf["file"],
        comment=COMMENT,
        index_col=0,
        parse_dates=True,
        na_values=conf["nodata"],
        names=conf["columns"],
        sep=DELIMITER,
        skiprows=conf["skiprows"],
        skip_blank_lines=True,
        usecols=range(len(conf["columns"])),
        low_memory=False,
    )
    try:
        df.index = pd.to_datetime(df.index)
    except ValueError as e:
        logger.info("\n" + conf["file"])
        logger.info("\nValueError:")
        logger.info(e)
        logger.info("\t\t> Trying pd.to_datetime with format=mixed")
        try:
            df.index = pd.to_datetime(df.index, format="mixed")
        except Exception as e:
            logger.info("\nDateParseError:")
            logger.info(e)
            logger.info(
                "\t\t> Trying again removing apostrophes in timestamp (old files format)"
            )
            df.index = pd.to_datetime(df.index.str.replace('"', ""))

    return df


def _parse_toa5(conf) -> pd.DataFrame:
    df = _parse_csv_default(conf)
    # TODO: Convert to xr.DataSet to allow for metadata enrichment
    try:
        meta = toa5.read_metadata(conf["file"])
        tao5_attrs = meta.get("attrs", {})
        tao5_attrs["file_format"] = tao5_attrs.pop("format")
    except Exception as e:
        logger.warning(f"Failed to enrich TOA5 metadata for {conf['file']}: {e}")
    return df


def load_data_file(conf: Dict) -> xr.Dataset:
    """Read L0 data file to xarray.Dataset using config dictionary and
    populate with initial metadata. The file type is detected automatically.

    Parameters
    ----------
    conf : dict
        Configuration parameters
    delimiter : str
    comment: str

    Returns
    -------
    ds : xr.Dataset
        L0 data
    """
    file_type = _detect_file_type(conf)
    logger.info(f"Detected L0 file type '{file_type}' for {conf.get('file')}")

    if file_type == "csv_v1":
        df = _parse_csv_v1(conf)
    elif file_type == "csv_default":
        df = _parse_csv_default(conf)
    elif file_type == "toa5":
        df = _parse_toa5(conf)
    else:
        raise ValueError(f"Unknown file type: {file_type}")

    df = _postprocess_dataframe(df, time_offset=conf.get("time_offset"))

    # Carry relevant metadata with ds
    ds = xr.Dataset.from_dataframe(df)
    ds.attrs["level"] = "L0"
    ds.attrs["detected_file_type"] = file_type
    ds.attrs["filename"] = Path(conf["file"]).name

    # populate meta with config keys
    skip = ["columns", "skiprows", "modem", "file", "conf", "nodata"]
    for k, v in conf.items():
        if k not in skip:
            ds.attrs[k] = v

    return ds


def load_data_files(config: Dict[str, Dict]) -> List[xr.Dataset]:
    """Load level 0 (L0) data from config mapping file names to configuration.

    Tries read_l0_file() using the config with msg_lat & msg_lon appended.
    If a pandas.errors.ParserError occurs due to mismatched columns, removes
    msg_lat & msg_lon from the config and tries again.

    Parameters
    ----------
    config : Dict[str, Dict]
        Configuration dictionary as returned by pypromice.io.load.getConfig

    Returns
    -------
    List[xr.Dataset]
        List of L0 datasets
    """
    ds_list: List[xr.Dataset] = []
    for k in config.keys():
        target = config[k]
        try:
            ds_list.append(load_data_file(target))
        except pd.errors.ParserError:
            for item in ["msg_lat", "msg_lon"]:
                if item in target["columns"]:
                    target["columns"].remove(item)
            ds_list.append(load_data_file(target))
        logger.info(f"L0 data successfully loaded from {k}")
    return ds_list


def _postprocess_dataframe(
    df: pd.DataFrame, time_offset: Optional[float] = None
) -> pd.DataFrame:
    """Apply common post-processing to parsed L0 dataframe."""
    if time_offset is not None:
        df.index = df.index + timedelta(hours=time_offset)
    # Drop SKIP columns
    for c in list(df.columns):
        if c.startswith("SKIP"):
            df.drop(columns=c, inplace=True)
    return df


def load_config(
    config_file: str | Path,
    inpath: str | Path,
    default_columns: Sequence[str] = ("msg_lat", "msg_lon"),
):
    """Load configuration from .toml file. PROMICE .toml files support defining
    features at the top level which apply to all nested properties, but do not
    overwrite nested properties if they are defined

    Parameters
    ----------
    config_file
        TOML file path
    inpath
        Input folder directory where L0 files can be found

    Returns
    -------
    conf : dict
        Configuration dictionary
    """
    config_file = Path(config_file)
    inpath = Path(inpath)

    conf = toml.load(config_file)  # Move all top level keys to nested properties,
    top = [
        _ for _ in conf.keys() if not type(conf[_]) is dict
    ]  # if they are not already defined in the nested properties
    subs = [
        _ for _ in conf.keys() if type(conf[_]) is dict
    ]  # Insert the section name (config_file) as a file property and config file
    for s in subs:
        for t in top:
            if t not in conf[s].keys():
                conf[s][t] = conf[t]

        conf[s]["conf"] = config_file.as_posix()
        conf[s]["file"] = os.path.join(inpath, s)
        conf[s]["columns"].extend(default_columns)

    for t in top:
        conf.pop(t)  # Delete all top level keys beause each file
    # should carry all properties with it
    for k in conf.keys():  # Check required fields are present
        for field in ["columns", "station_id", "format", "skiprows"]:
            assert field in conf[k].keys(), field + " not in config keys"
    return conf



__all__ = ["decode_and_convert",  "filter",
           "decode", "convert_from_degrees_and_decimal_minutes",
           "convert_from_decimal_minutes"]
import re
import xarray as xr
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression

import logging
logger = logging.getLogger(__name__)

def decode_and_convert(gps_lat: xr.DataArray,
                       gps_lon: xr.DataArray,
                       gps_time: xr.DataArray,
                       latitude: float,
                       longitude: float
) -> tuple[xr.DataArray,xr.DataArray,xr.DataArray]:
    """Decode and convert GPS latitude, longtitude and time values."flag_decimal_minutes",
           "flag_for_decoding"
    Decoding is performed if values are detected as string types.
    Conversion consists of transforming to decimal degrees (DD),
    from either decimal minutes (mm.mmmmm) or degrees and
    decimal minutes (ddmm.mmmm)

    Parameters
    ----------
    gps_lat : `xr.DataArray`
        GPS latitude
    gps_lon : `xr.DataArray`
        GPS longitude
    gps_time : `xr.DataArray`
        GPS time

    Returns
    -------
    gps_lat : `xr.DataArray`
        Decoded and converted GPS latitude
    gps_lon : `xr.DataArray`
        Decoded and converted GPS longitude
    gps_time : `xr.DataArray`
        Decoded and converted GPS time
    """
    # Retain GPS array attributes
    lat_attrs = gps_lat.attrs
    lon_attrs = gps_lon.attrs
    time_attrs = gps_time.attrs

    # Decode GPS information if array is an object array
    if gps_lat.dtype.kind == "O":
        lat, lon, time = decode(gps_lat, gps_lon, gps_time)
        if lat is None:
            logger.warning("GPS decoding failed, skipping this routine.")
        else:
            gps_lat, gps_lon, gps_time = lat, lon, time

    # Reformat values to numeric
    gps_lat.values = pd.to_numeric(gps_lat, errors='coerce')
    gps_lon.values = pd.to_numeric(gps_lon, errors='coerce')
    gps_time.values = pd.to_numeric(gps_time, errors='coerce')

    # Convert GPS positions to decimal degrees
    if np.any((gps_lat <= 90) & (gps_lat > 0)):
        gps_lat = convert_from_decimal_minutes(gps_lat, latitude)
        gps_lon = convert_from_decimal_minutes(gps_lon, longitude)
    else:
        gps_lat = convert_from_degrees_and_decimal_minutes(gps_lat)
        gps_lon = convert_from_degrees_and_decimal_minutes(gps_lon)

    # Reassign GPS array attributes
    gps_lat.attrs = lat_attrs
    gps_lon.attrs = lon_attrs
    gps_time.attrs = time_attrs

    return gps_lat, gps_lon, gps_time


def filter(gps_lat: xr.DataArray,
           gps_lon: xr.DataArray,
           gps_alt: xr.DataArray
) -> tuple[xr.DataArray, xr.DataArray, xr.DataArray]:
    """ Filter GPS latitude, longitude and altitude based on the difference
    to a baseline elevation. The baseline elevation is a gap-filled monthly
    median elevation based on the inputted GPS altitude.

    Parameters
    ----------
    gps_lat : xr.DataArray
        GPS latitude
    gps_lon : xr.DataArray
        GPS longitude
    gps_alt : xr.DataArray
        GPS altitude values with a time dimension

    Returns
    ----------
    gps_lat_filtered : xr.DataArray
        Filtered latitude values
    gps_lon_filtered : xr.DataArray
        Filtered longitude values
    gps_alt_filtered : xr.DataArray
        Filtered altitude values
    """
    # Get altitude monthly median (at month start)
    # This will serve as baseline elevations for filtering
    ser = gps_alt.to_series()
    monthly_median = ser.resample("MS").median()
    baseline_elevation = (
        monthly_median
        .reindex(ser.index, method="nearest")
        .ffill()
        .bfill()
    )

    # Produce conditional mask
    mask = (np.abs(gps_alt - baseline_elevation) < 100) | gps_alt.isnull()

    # Apply mask
    gps_lat_filtered = gps_lat.where(mask)
    gps_lon_filtered = gps_lon.where(mask)
    gps_alt_filtered = gps_alt.where(mask)

    return gps_lat_filtered, gps_lon_filtered, gps_alt_filtered


def convert_from_degrees_and_decimal_minutes(gps):
    """Convert positions (i.e. latitude, longitude) from degrees
    and decimal minutes (ddmm.mmmm) to decimal degree values (DD)"""
    return np.floor(gps / 100) + (gps / 100 - np.floor(gps / 100)) * 100 / 60


def convert_from_decimal_minutes(gps: xr.DataArray, pos: float
) -> xr.DataArray:
    """Convert decimal minutes (mm.mmmmm) to decimal degree
    values (DD), using a predefined position to append values to.
    Needed in the case of PROMICE v1 stations, where logger
    programs saved positions only in decimal minutes."""
    new_gps = np.sign(pos) * (gps + 100 * np.floor(np.abs(pos)))
    return convert_from_degrees_and_decimal_minutes(new_gps)


def decode(gps_lat: xr.DataArray,
           gps_lon: xr.DataArray,
           gps_time: xr.DataArray
) -> tuple[xr.DataArray,xr.DataArray,xr.DataArray]:
    """Decode GPS information. This should be applied if gps information
    consists of strings and not float values. GPS information is returned in
    decimal degrees (ddmm.mmmm) format.

    Parameters
    ----------
    gps_lat : `xr.DataArray`
        GPS latitude
    gps_lon : `xr.DataArray`
        GPS longitude
    gps_time : `xr.DataArray`
        GPS time

    Returns
    -------
    new_lat : `xr.DataArray`
        Decoded GPS latitude
    new_lon : `xr.DataArray`
        Decoded GPS longitude
    new_time : `xr.DataArray`
        Decoded GPS time
    """
    # Pick the first non-null sample safely and detect decoding format
    non_null = gps_lat.dropna(dim='time').values
    sample_value = str(non_null[0])

    try:
        # Object decoding
        if "NH" in sample_value:
            new_lat = gps_object_decoder(gps_lat)
            new_lon = gps_object_decoder(gps_lon)
            new_time = gps_object_decoder(gps_time)
            return new_lat, new_lon, new_time

        # L-string decoding
        elif "L" in sample_value:
            logger.info("Found 'L' in GPS string; applying decode + scaling.")
            new_lat = gps_l_string_decoder(gps_lat)
            new_lon = gps_l_string_decoder(gps_lon)
            new_time = gps_object_decoder(gps_time)
            return new_lat, new_lon, new_time

        # Unknown format, attempt to decode
        else:
            logger.info("Unknown GPS string format; attempting generic decode.")
            new_lat = gps_object_decoder(gps_lat)
            new_lon = gps_object_decoder(gps_lon)
            new_time = gps_object_decoder(gps_time)
            return new_lat, new_lon, new_time

    except Exception as e:
        logger.error(f"Failed to decode GPS data: {e!r} "
                     f"(dtype={gps_lat.dtype})")
        return None, None, None


def gps_object_decoder(gps : xr.DataArray) -> xr.DataArray:
    """GPS decoder for object array formatting. For example, PROMICE v2
    stations should send information as 'NH6429.01544,WH04932.86061'
    original formatting (NUK_L 2022); PROMICE v3 stations should send
    coordinates as '6430,4916' (NUK_Uv3); and GC-Net stations should
    send coordinates as '6628.93936',04617.59187' (DY2)"""
    str2nums = [re.findall(r"[-+]?\d*\.\d+|\d+", _) if isinstance(_, str) else [np.nan] for _ in gps.values]
    gps[:] = pd.DataFrame(str2nums).astype(float).T.values[0]
    gps = gps.astype(float)
    return gps


def gps_l_string_decoder(gps : xr.DataArray) -> xr.DataArray:
    """GPS L-string decoder"""
    # Convert from object array
    gps = gps_object_decoder(gps)

    # Convert from integer-like values to degrees
    gps = gps/100000
    return gps
